/*
    libutf8filename -- transparently convert utf8 filenames into 8 bit ones
    Copyright (C) 2003  Egmont Koblinger <egmont@uhulinux.hu>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <fcntl.h>
#include <utime.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <locale.h>
#include <langinfo.h>
#include <iconv.h>
#include <limits.h>

static locale_t locale = 0;
static char *charset;
static iconv_t ic_local2utf;
static iconv_t ic_utf2local;

static struct dirent *(*next_readdir) (DIR *dir);
static struct dirent64 *(*next_readdir64) (DIR *dir);
static char *(*next_getcwd) (char *, size_t);
static int (*next_readlink) (const char *, char *, size_t);
static int (*next_chdir) (const char *);
static int (*next_open) (const char *, int, ...);
static int (*next_open64) (const char *, int, ...);
static int (*next_creat) (const char *, mode_t);
static int (*next_creat64) (const char *, mode_t);
static DIR *(*next_opendir) (const char *);
static int (*next_access) (const char *, int);
static int (*next___xstat) (int, const char *, struct stat *);
static int (*next___xstat64) (int, const char *, struct stat64 *);
static int (*next___lxstat) (int, const char *, struct stat *);
static int (*next___lxstat64) (int, const char *, struct stat64 *);
static int (*next_statfs) (const char *, struct statfs *);
static int (*next_chmod) (const char *, mode_t);
static int (*next_chown) (const char *, uid_t, gid_t);
static int (*next_lchown) (const char *, uid_t, gid_t);
static int (*next_utime) (const char *, const struct utimbuf *);
static int (*next_utimes) (const char *, const struct timeval *);
static int (*next_lutimes) (const char *, const struct timeval *);
static int (*next_link) (const char *, const char *);
static int (*next_symlink) (const char *, const char *);
static int (*next_mkdir) (const char *, mode_t);
static int (*next_mknod) (const char *, mode_t, dev_t);
static int (*next_rename) (const char *, const char *);
static int (*next_unlink) (const char *);
static int (*next_rmdir) (const char *);
static int (*next_truncate) (const char *, off_t);

static int do_wrap = 0;

void sortdir_init (void) __attribute((constructor));
void sortdir_init (void)
{
	locale = newlocale(LC_ALL_MASK, "", NULL);
	charset = nl_langinfo_l(CODESET, locale);
	ic_local2utf = iconv_open("UTF-8", charset);
	ic_utf2local = iconv_open(charset, "UTF-8");

	next_readdir    = dlsym(RTLD_NEXT, "readdir");
	next_readdir64  = dlsym(RTLD_NEXT, "readdir64");
	next_getcwd     = dlsym(RTLD_NEXT, "getcwd");
	next_readlink   = dlsym(RTLD_NEXT, "readlink");
	next_chdir      = dlsym(RTLD_NEXT, "chdir");
	next_open       = dlsym(RTLD_NEXT, "open");
	next_open64     = dlsym(RTLD_NEXT, "open64");
	next_creat      = dlsym(RTLD_NEXT, "creat");
	next_creat64    = dlsym(RTLD_NEXT, "creat64");
	next_opendir    = dlsym(RTLD_NEXT, "opendir");
	next_access     = dlsym(RTLD_NEXT, "access");
	next___xstat    = dlsym(RTLD_NEXT, "__xstat");
	next___xstat64  = dlsym(RTLD_NEXT, "__xstat64");
	next___lxstat   = dlsym(RTLD_NEXT, "__lxstat");
	next___lxstat64 = dlsym(RTLD_NEXT, "__lxstat64");
	next_statfs     = dlsym(RTLD_NEXT, "statfs");
	next_chmod      = dlsym(RTLD_NEXT, "chmod");
	next_chown      = dlsym(RTLD_NEXT, "chown");
	next_lchown     = dlsym(RTLD_NEXT, "lchown");
	next_utime      = dlsym(RTLD_NEXT, "utime");
	next_utimes     = dlsym(RTLD_NEXT, "utimes");
	next_lutimes    = dlsym(RTLD_NEXT, "lutimes");
	next_link       = dlsym(RTLD_NEXT, "link");
	next_symlink    = dlsym(RTLD_NEXT, "symlink");
	next_mknod      = dlsym(RTLD_NEXT, "mknod");
	next_mkdir      = dlsym(RTLD_NEXT, "mkdir");
	next_rename     = dlsym(RTLD_NEXT, "rename");
	next_unlink     = dlsym(RTLD_NEXT, "unlink");
	next_rmdir      = dlsym(RTLD_NEXT, "rmdir");
	next_truncate   = dlsym(RTLD_NEXT, "truncate");

	do_wrap = 1;
}

/*--------------------------------------------------------------------------*/

static int utf2local (const char *utf, char *local)
{
	size_t i;
	size_t inbytesleft = strlen(utf);
	size_t outbytesleft = PATH_MAX;
	iconv(ic_utf2local, NULL, 0, NULL, 0);
	i = iconv(ic_utf2local, &utf, &inbytesleft, &local, &outbytesleft);
	if (i == (size_t)(-1)) return -1;
	*local = 0;
	return 0;
}

static int local2utf (const char *local, char *utf)
{
	size_t i;
	size_t inbytesleft = strlen(local);
	size_t outbytesleft = PATH_MAX;
	iconv(ic_local2utf, NULL, 0, NULL, 0);
	i = iconv(ic_local2utf, &local, &inbytesleft, &utf, &outbytesleft);
	if (i == (size_t)(-1)) return -1;
	*utf = 0;
	return 0;
}

/*--------------------------------------------------------------------------*/

struct dirent *readdir (DIR *dir)
{
	struct dirent *dirent;
	char local[NAME_MAX + 1];
	while (1) {
		dirent = next_readdir(dir);
		if (dirent == NULL) return NULL;
		utf2local(dirent->d_name, local);
		strcpy(dirent->d_name, local);
#ifdef _DIRENT_HAVE_D_NAMLEN
		d->namlen = strlen(dirent->d_name);
#endif
		return dirent;
	}
}

struct dirent64 *readdir64 (DIR *dir)
{
	struct dirent64 *dirent;
	char local[NAME_MAX + 1];
	while (1) {
		dirent = next_readdir64(dir);
		if (dirent == NULL) return NULL;
		utf2local(dirent->d_name, local);
		strcpy(dirent->d_name, local);
#ifdef _DIRENT_HAVE_D_NAMLEN
		d->namlen = strlen(dirent->d_name);
#endif
		return dirent;
	}
}

char *getcwd (char *buf, size_t size)
{
	char *c;
	int i;
	char utf[PATH_MAX + 1];
	char local[PATH_MAX + 1];
	c = next_getcwd(utf, PATH_MAX);
	if (c != utf) return c;
	utf2local(utf, local);
	i = strlen(local);
	if (i > size - 1) i = size - 1;
	memcpy(buf, local, i + 1);
	return buf;
}

/*--------------------------------------------------------------------------*/

int readlink (const char *local1, char *buf, size_t bufsiz)
{
	int i;
	char utf1[PATH_MAX + 1];
	char utf2[PATH_MAX + 1];
	char local2[PATH_MAX + 1];
	local2utf(local1, utf1);
	i = next_readlink(utf1, utf2, PATH_MAX);
	if (i < 0) return i;
	utf2[i] = 0;
	utf2local(utf2, local2);
	i = strlen(local2);
	if (i > bufsiz) i = bufsiz;
	memcpy(buf, local2, i);
	return i;
}

/*--------------------------------------------------------------------------*/

int chdir (const char *local)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_chdir(utf);
}

int open (const char *local, int flags, ...)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	if (flags & O_CREAT) {
		va_list ap;
		mode_t mode;
		va_start(ap, flags);
		mode = va_arg(ap, mode_t);
		va_end(ap);
		return next_open(utf, flags, mode);
	} else {
		return next_open(utf, flags);
	}
}

int open64 (const char *local, int flags, ...)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	if (flags & O_CREAT) {
		va_list ap;
		mode_t mode;
		va_start(ap, flags);
		mode = va_arg(ap, mode_t);
		va_end(ap);
		return next_open64(utf, flags, mode);
	} else {
		return next_open64(utf, flags);
	}
}

int creat (const char *local, mode_t mode)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_creat(utf, mode);
}

int creat64 (const char *local, mode_t mode)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_creat64(utf, mode);
}

DIR *opendir (const char *local)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_opendir(utf);
}

int access (const char *local, int mode)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_access(utf, mode);
}

int __xstat (int ver, const char *local, struct stat *st)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next___xstat(ver, utf, st);
}

int __xstat64 (int ver, const char *local, struct stat64 *st)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next___xstat64(ver, utf, st);
}

int __lxstat (int ver, const char *local, struct stat *st)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next___lxstat(ver, utf, st);
}

int __lxstat64 (int ver, const char *local, struct stat64 *st)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next___lxstat64(ver, utf, st);
}

int statfs (const char *local, struct statfs *buf)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_statfs(utf, buf);
}

int chmod (const char *local, mode_t mode)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_chmod(utf, mode);
}

int chown (const char *local, uid_t uid, gid_t gid)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_chown(utf, uid, gid);
}

int lchown (const char *local, uid_t uid, gid_t gid)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_lchown(utf, uid, gid);
}

int utime (const char *local, const struct utimbuf *buf)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_utime(utf, buf);
}

int utimes (const char *local, const struct timeval *tvp)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_utimes(utf, tvp);
}

int lutimes (const char *local, const struct timeval *tvp)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_lutimes(utf, tvp);
}

int link (const char *local1, const char *local2)
{
	char utf1[PATH_MAX + 1];
	char utf2[PATH_MAX + 1];
	local2utf(local1, utf1);
	local2utf(local2, utf2);
	return next_link(utf1, utf2);
}

int symlink (const char *local1, const char *local2)
{
	char utf1[PATH_MAX + 1];
	char utf2[PATH_MAX + 1];
	local2utf(local1, utf1);
	local2utf(local2, utf2);
	return next_symlink(utf1, utf2);
}

int mkdir (const char *local, mode_t mode)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_mkdir(utf, mode);
}

int mknod (const char *local, mode_t mode, dev_t dev)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_mknod(utf, mode, dev);
}

int rename (const char *local1, const char *local2)
{
	char utf1[PATH_MAX + 1];
	char utf2[PATH_MAX + 1];
	local2utf(local1, utf1);
	local2utf(local2, utf2);
	return next_rename(utf1, utf2);
}

int unlink (const char *local)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_unlink(utf);
}

int rmdir (const char *local)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_rmdir(utf);
}

int truncate (const char *local, off_t off)
{
	char utf[PATH_MAX + 1];
	local2utf(local, utf);
	return next_truncate(utf, off);
}

/*--------------------------------------------------------------------------*/
